/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  SplitFieldIntoNrrdData.cc:  Convert a Nrrd to a Field
 *
 *  Written by:
 *   David Weinstein
 *   School of Computing
 *   University of Utah
 *   February 2001
 *
 */

#include <Core/Algorithms/Converter/ConvertToNrrd.h>

#include <Dataflow/Network/Ports/NrrdPort.h>
#include <Dataflow/Network/Ports/FieldPort.h>
#include <Dataflow/Network/Module.h>

namespace SCITeem {

using namespace SCIRun;

class SplitFieldIntoNrrdData : public Module {
  public:
    SplitFieldIntoNrrdData(GuiContext *ctx);
    virtual ~SplitFieldIntoNrrdData() {}
    virtual void execute();

  private:
    GuiString gui_label_;
    
    SCIRunAlgo::ConvertToNrrdAlgo algo_;
};

} // end namespace SCITeem

using namespace SCITeem;
DECLARE_MAKER(SplitFieldIntoNrrdData)


SplitFieldIntoNrrdData::SplitFieldIntoNrrdData(GuiContext *ctx):
  Module("SplitFieldIntoNrrdData", ctx, Filter, "Converters", "Teem"),
  gui_label_(get_ctx()->subVar("label"), "unknown")
{
  algo_.set_progress_reporter(this);
}

void
SplitFieldIntoNrrdData::execute()
{
  FieldHandle field_handle; 
  if (!get_input_handle("Field", field_handle)) return;

  // Just data for lattices, data and points for structured, all for rest.
  bool compute_points_p = true;
  bool compute_connects_p = true;
  bool compute_data_p = true;

  if (!field_handle->vmesh()->is_editable())
  {
    remark("Not computing connections for non-editable mesh type.");
    compute_connects_p = false;
  }
  
  if (field_handle->vfield()->basis_order() == -1)
  {
    remark("No data in input field.");
    compute_data_p = false;
  }

  if (inputs_changed_ ||
      gui_label_.changed() ||
      (compute_points_p && !oport_cached("Points")) ||
      (compute_connects_p && !oport_cached("Connections")) ||
      (compute_data_p && !oport_cached("Data")) )
  {
    NrrdDataHandle points_handle, connect_handle, data_handle;

    algo_.set_bool("build_points",compute_points_p);
    algo_.set_bool("build_connections",compute_connects_p);
    algo_.set_bool("build_data",compute_data_p);
    algo_.set_string("data_label",gui_label_.get());
    if(!(algo_.run(field_handle,points_handle, connect_handle, data_handle))) return;

    // Set the Nrrd names and send them.
    std::string property;
    std::string nrrd_name = "Unknown";
    if (field_handle->get_property( "name", property ) && property != "Unknown") 
      nrrd_name = property;

    if (points_handle.get_rep()) 
    {
      points_handle->set_property("Name", nrrd_name + "-Points", false);
      send_output_handle("Points", points_handle);
    }

    if (connect_handle.get_rep()) 
    {
      connect_handle->set_property("Name", nrrd_name + "-Connectivity", false);
      send_output_handle("Connections", connect_handle);
    }

    if (data_handle.get_rep()) 
    {
      data_handle->set_property("Name", nrrd_name + "-Data", false);
      send_output_handle("Data", data_handle);
    }
  }
}

